@@ -22,10 +22,10 @@ except ImportError:
 from .S3 import S3
 from .Config import Config
 from .Exceptions import *
-from .Utils import (getTreeFromXml, appendXmlTextNode, getDictFromTree,
-                    dateS3toPython, getBucketFromHostname,
-                    getHostnameFromBucket, deunicodise, urlencode_string,
-                    convertHeaderTupleListToDict, encode_to_s3, decode_from_s3)
+from .BaseUtils import (getTreeFromXml, appendXmlTextNode, getDictFromTree,
+                        dateS3toPython, encode_to_s3, decode_from_s3)
+from .Utils import (getBucketFromHostname, getHostnameFromBucket, deunicodise,
+                    urlencode_string, convertHeaderTupleListToDict)
 from .Crypto import sign_string_v2
 from .S3Uri import S3Uri, S3UriS3
 from .ConnMan import ConnMan
@@ -483,7 +483,7 @@ class CloudFront(object):
                     fp.write(deunicodise("\n".join(paths)+"\n"))
                 warning("Request to invalidate %d paths (max 999 supported)" % len(paths))
                 warning("All the paths are now saved in: %s" % tmp_filename)
-            except:
+            except Exception:
                 pass
             raise ParameterError("Too many paths to invalidate")
 
@@ -622,7 +622,7 @@ class CloudFront(object):
                     #       do this since S3 buckets that are set up as websites use custom origins.
                     #       Thankfully, the custom origin URLs they use start with the URL of the
                     #       S3 bucket. Here, we make use this naming convention to support this use case.
-                    distListIndex = getBucketFromHostname(d.info['CustomOrigin']['DNSName'])[0];
+                    distListIndex = getBucketFromHostname(d.info['CustomOrigin']['DNSName'])[0]
                     distListIndex = distListIndex[:len(uri.bucket())]
                 else:
                     # Aral: I'm not sure when this condition will be reached, but keeping it in there.
@@ -803,7 +803,7 @@ class Cmd(object):
                 try:
                     for i in inval_list['inval_list'].info['InvalidationSummary']:
                         requests.append("/".join(["cf:/", cfuri.dist_id(), i["Id"]]))
-                except:
+                except Exception:
                     continue
         for req in requests:
             cfuri = S3Uri(req)
