@@ -1,18 +1,36 @@
+# -*- coding: utf-8 -*-
+
 ## Amazon S3 manager - Exceptions library
 ## Author: Michal Ludvig <michal@logix.cz>
 ##         http://www.logix.cz/michal
 ## License: GPL Version 2
 ## Copyright: TGRMN Software and contributors
 
-from Utils import getTreeFromXml, unicodise, deunicodise
-from logging import debug, info, warning, error
-import ExitCodes
+from __future__ import absolute_import
+
+from logging import debug, error
+import sys
+import S3.BaseUtils
+import S3.Utils
+from . import ExitCodes
+
+if sys.version_info >= (3, 0):
+    PY3 = True
+    # In python 3, unicode -> str, and str -> bytes
+    unicode = str
+else:
+    PY3 = False
+
+## External exceptions
+
+from ssl import SSLError as S3SSLError
 
 try:
-    import xml.etree.ElementTree as ET
+    from ssl import CertificateError as S3SSLCertificateError
 except ImportError:
-    # xml.etree.ElementTree was only added in python 2.5
-    import elementtree.ElementTree as ET
+    class S3SSLCertificateError(Exception):
+        pass
+
 
 try:
     from xml.etree.ElementTree import ParseError as XmlParseError
@@ -20,14 +38,20 @@ except ImportError:
     # ParseError was only added in python2.7, before ET was raising ExpatError
     from xml.parsers.expat import ExpatError as XmlParseError
 
+
+## s3cmd exceptions
+
 class S3Exception(Exception):
-    def __init__(self, message = ""):
-        self.message = unicodise(message)
+    def __init__(self, message=""):
+        self.message = S3.Utils.unicodise(message)
 
     def __str__(self):
-        ## Call unicode(self) instead of self.message because
-        ## __unicode__() method could be overriden in subclasses!
-        return deunicodise(unicode(self))
+        ## Don't return self.message directly because
+        ## __unicode__() method could be overridden in subclasses!
+        if PY3:
+            return self.__unicode__()
+        else:
+            return S3.Utils.deunicodise(self.__unicode__())
 
     def __unicode__(self):
         return self.message
@@ -35,6 +59,7 @@ class S3Exception(Exception):
     ## (Base)Exception.message has been deprecated in Python 2.6
     def _get_message(self):
         return self._message
+
     def _set_message(self, message):
         self._message = message
     message = property(_get_message, _set_message)
@@ -45,23 +70,23 @@ class S3Error (S3Exception):
         self.status = response["status"]
         self.reason = response["reason"]
         self.info = {
-            "Code" : "",
-            "Message" : "",
-            "Resource" : ""
+            "Code": "",
+            "Message": "",
+            "Resource": ""
         }
         debug("S3Error: %s (%s)" % (self.status, self.reason))
-        if response.has_key("headers"):
+        if "headers" in response:
             for header in response["headers"]:
                 debug("HttpHeader: %s: %s" % (header, response["headers"][header]))
-        if response.has_key("data") and response["data"]:
+        if "data" in response and response["data"]:
             try:
-                tree = getTreeFromXml(response["data"])
+                tree = S3.BaseUtils.getTreeFromXml(response["data"])
             except XmlParseError:
                 debug("Not an XML response")
             else:
                 try:
                     self.info.update(self.parse_error_xml(tree))
-                except Exception, e:
+                except Exception as e:
                     error("Error parsing xml: %s.  ErrorXML: %s" % (e, response["data"]))
 
         self.code = self.info["Code"]
@@ -70,15 +95,16 @@ class S3Error (S3Exception):
 
     def __unicode__(self):
         retval = u"%d " % (self.status)
-        retval += (u"(%s)" % (self.info.has_key("Code") and self.info["Code"] or self.reason))
-        if self.info.has_key("Message"):
-            retval += (u": %s" % self.info["Message"])
+        retval += (u"(%s)" % ("Code" in self.info and self.info["Code"] or self.reason))
+        error_msg = self.info.get("Message")
+        if error_msg:
+            retval += (u": %s" % error_msg)
         return retval
 
     def get_error_code(self):
         if self.status in [301, 307]:
             return ExitCodes.EX_SERVERMOVED
-        elif self.status in [400, 405, 411, 416, 501]:
+        elif self.status in [400, 405, 411, 416, 417, 501, 504]:
             return ExitCodes.EX_SERVERERROR
         elif self.status == 403:
             return ExitCodes.EX_ACCESSDENIED
@@ -90,7 +116,7 @@ class S3Error (S3Exception):
             return ExitCodes.EX_PRECONDITION
         elif self.status == 500:
             return ExitCodes.EX_SOFTWARE
-        elif self.status == 503:
+        elif self.status in [429, 503]:
             return ExitCodes.EX_SERVICE
         else:
             return ExitCodes.EX_SOFTWARE
@@ -102,7 +128,7 @@ class S3Error (S3Exception):
         if not error_node.tag == "Error":
             error_node = tree.find(".//Error")
         if error_node is not None:
-            for child in error_node.getchildren():
+            for child in error_node:
                 if child.text != "":
                     debug("ErrorXML: " + child.tag + ": " + repr(child.text))
                     info[child.tag] = child.text
